/**
 TULCS.java
 Created on February 6, 2007, 12:37 PM
 
 
 SequenceMiner: multiple anomaly classifier
 
 @author: Eugene Turkov, {eturkov@email.arc.nasa.gov}
 @version: 1.0
 
 NASA Ames Research Center
 Computational Sciences Division
 Intelligent Data Understanding Group (IDU)
 Moffett Field, CA 94035
 */

package turkov.math.algorithms;

public class TULCS
{
    public static int getLengthOfLongestCommonSubsequence(Object[] a, Object[] b)
    {
        int n = a.length;
        int m = b.length;
        int current[] = new int[m+1];
        int previous[] = new int[m+1];
        int row, column;
        
        // It is important to use <=, not <.  The next two for-loops are initialization

        
        // This is the main dynamic programming loop that computes the score and
        // backtracking arrays.
        for(row = 1; row <= n; ++row)
        {
            for(column = 1; column <= m; ++column)
            {                
                if( a[row-1].equals(b[column-1]))
                    current[column] = previous[column - 1] + 1;
                
                else
                    current[column] = previous[column-1];
                
                if( previous[column] >= current[column] )
                    current[column] = previous[column];
                
                if( current[column-1] >= current[column] )
                    current[column] = current[column-1];
            }
            
            previous = current;
            current = new int[previous.length];
        }
        
        return previous[m];
    }
    
    public static Object[] getLongestCommonSubsequence(Object[] a, Object[] b)
    {
        int n = a.length;
        int m = b.length;
        int S[][] = new int[n+1][m+1];
        int R[][] = new int[n+1][m+1];
        int ii, jj;
        
        // It is important to use <=, not <.  The next two for-loops are initialization
        for(ii = 0; ii <= n; ++ii)
        {
            S[ii][0] = 0;
            R[ii][0] = UP;
        }
        for(jj = 0; jj <= m; ++jj)
        {
            S[0][jj] = 0;
            R[0][jj] = LEFT;
        }
        
        // This is the main dynamic programming loop that computes the score and
        // backtracking arrays.
        for(ii = 1; ii <= n; ++ii)
        {
            for(jj = 1; jj <= m; ++jj)
            {
                
                if( a[ii-1].equals(b[jj-1]))
                {
                    S[ii][jj] = S[ii-1][jj-1] + 1;
                    R[ii][jj] = UP_AND_LEFT;
                }
                
                else
                {
                    S[ii][jj] = S[ii-1][jj-1] + 0;
                    R[ii][jj] = NEITHER;
                }
                
                if( S[ii-1][jj] >= S[ii][jj] )
                {
                    S[ii][jj] = S[ii-1][jj];
                    R[ii][jj] = UP;
                }
                
                if( S[ii][jj-1] >= S[ii][jj] )
                {
                    S[ii][jj] = S[ii][jj-1];
                    R[ii][jj] = LEFT;
                }
            }
        }
        
        // The length of the longest substring is S[n][m]
        ii = n;
        jj = m;
        int pos = S[ii][jj] - 1;
        Object[] lcs = new Object[ pos+1 ];
        
        // Trace the backtracking matrix.
        while( ii > 0 || jj > 0 )
        {
            if( R[ii][jj] == UP_AND_LEFT )
            {
                ii--;
                jj--;
                lcs[pos--] = a[ii];
            }
            
            else if( R[ii][jj] == UP )
                ii--;
            
            else if( R[ii][jj] == LEFT )
                jj--;
        }
        
        return lcs;
    }

    private static final short NEITHER     = 0;
    private static final short UP          = 1;
    private static final short LEFT        = 2;
    private static final short UP_AND_LEFT = 3;  
}
